package aligner;

import dna.AminoAcid;
import shared.Tools;

/**
 * @author Brian Bushnell
 * @contributor Isla
 * @date June 3, 2025
 */
public class Query {

	public Query(String name_, long nid, byte[] bases_, byte[] quals_) {
		name=name_;
		numericID=nid;
		bases=bases_;
		quals=quals_;
		
		rbases=AminoAcid.reverseComplementBases(bases);
		int[][] index=makeIndex(bases);
		kmers=index[0];
		rkmers=index[1];
		validKmers=(kmers==null ? 0 : Tools.countGreaterThan(kmers, -1));
		minHits=(mhc==null ? 0 : mhc.minHits(validKmers));
		maxClips=(maxClip<1 ? (int)maxClip*bases.length : (int)maxClip);
	}
	
	public int length() {return bases.length;}

	private static int[][] makeIndex(byte[] sequence){
		if(sequence.length<k || !indexKmers){return blankIndex;}
		assert(indexKmers);

		final int shift=2*k;
		final int shift2=shift-2;
		final int mask=~((-1)<<shift);
		int kmer=0, rkmer=0;
		int len=0;
		
		int[][] ret=new int[2][sequence.length-k+1];
		int kmerCount=0;

		for(int i=0, idx=-k+1; i<sequence.length; i++, idx++){
			final byte b=sequence[i];
			final int x=AminoAcid.baseToNumber[b];
			final int x2=AminoAcid.baseToComplementNumber[b];
			kmer=((kmer<<2)|x)&mask;
			rkmer=((rkmer>>>2)|(x2<<shift2))&mask;

			if(x<0){len=0; rkmer=0;}else{len++;}
			if(idx>=0) {
				if(len>=k && !AminoAcid.isHomopolymer(kmer, k, blacklistRepeatLength)){
					// Apply wildcard mask and store canonical k-mer
					ret[0][idx]=(kmer&midMask);
					ret[1][idx]=(rkmer&midMask);
					kmerCount++;
				}else {
					ret[0][idx]=ret[1][idx]=-1;
				}
			}
		}
		Tools.reverseInPlace(ret[1]);
		return kmerCount<1 ? blankIndex : ret;
	}

	public static int makeMidMask(int k, int maskLen) {
		if(maskLen<1) {return -1;}
		int bitsPerBase=2;
		assert(k>maskLen+1);
		int bits=maskLen*bitsPerBase;
		int shift=((k-maskLen)/2)*bitsPerBase;
		int middleMask=~((~((-1)<<bits))<<shift);
		return middleMask;
	}
	
	public static void setMode(int k_, int midMaskLen_, boolean indexKmers_) {
		setK(k_);
		setMidMaskLen(midMaskLen_);
		indexKmers=indexKmers_;
		assert(!indexKmers || (k>0 && k<16 && midMaskLen<k+1));
	}
	
	private static void setK(int x) {
		k=Math.max(x, 0);
		assert(k<16);
		indexKmers=k>0;
	}
	
	private static void setMidMaskLen(int x) {
		midMaskLen=Math.max(x, 0);
		assert(midMaskLen>=0 && (midMaskLen<1 || midMaskLen<k-1));
		midMask=makeMidMask(k, midMaskLen);
	}

	public final String name;
	public final long numericID;
	public final byte[] bases;
	public final byte[] rbases;
	public final byte[] quals;
	public final int[] kmers;
	public final int[] rkmers;
	public final int validKmers;
	public final int minHits;
	public final int maxClips;

	private static int k=11;
	private static int midMaskLen=1;
	public static int midMask=makeMidMask(k, midMaskLen);
	public static boolean indexKmers=true;
	public static int blacklistRepeatLength=2;
	private static final int[][] blankIndex=new int[2][];
	public static MinHitsCalculator mhc;
	public static float maxClip=0.25f;
	
}